"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const pg_1 = require("pg");
const AbstractSqlConnection_1 = require("./AbstractSqlConnection");
const TableCompiler_PG = require('knex/lib/dialects/postgres/schema/tablecompiler.js');
const TableCompiler = require('knex/lib/schema/tablecompiler.js');
class PostgreSqlConnection extends AbstractSqlConnection_1.AbstractSqlConnection {
    async connect() {
        this.patchKnex();
        this.client = this.createKnexClient('pg');
    }
    getDefaultClientUrl() {
        return 'postgresql://postgres@127.0.0.1:5432';
    }
    getConnectionOptions() {
        const ret = super.getConnectionOptions();
        pg_1.types.setTypeParser(1700, str => parseFloat(str));
        if (this.config.get('forceUtcTimezone')) {
            [1082, 1083, 1114].forEach(oid => pg_1.types.setTypeParser(oid, str => new Date(str + 'Z'))); // date, time, timestamp types
            pg_1.defaults.parseInputDatesAsUTC = true;
        }
        return ret;
    }
    transformRawResult(res, method) {
        if (method === 'get') {
            return res.rows[0];
        }
        if (method === 'all') {
            return res.rows;
        }
        return {
            affectedRows: res.rowCount,
            insertId: res.rows[0] ? res.rows[0].id : 0,
            row: res.rows[0],
        };
    }
    /**
     * monkey patch knex' postgres dialect so it correctly handles column updates (especially enums)
     */
    patchKnex() {
        const that = this;
        TableCompiler_PG.prototype.addColumns = function (columns, prefix, colCompilers) {
            if (prefix !== this.alterColumnsPrefix) {
                // base class implementation for normal add
                return TableCompiler.prototype.addColumns.call(this, columns, prefix);
            }
            // alter columns
            for (const col of colCompilers) {
                that.addColumn.call(this, col, that);
            }
        };
    }
    addColumn(col, that) {
        const quotedTableName = this.tableName();
        const type = col.getColumnType();
        const colName = this.client.wrapIdentifier(col.getColumnName(), col.columnBuilder.queryContext());
        const constraintName = `${this.tableNameRaw}_${col.getColumnName()}_check`;
        this.pushQuery({ sql: `alter table ${quotedTableName} drop constraint if exists "${constraintName}"`, bindings: [] });
        if (col.type === 'enu') {
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} type text using (${colName}::text)`, bindings: [] });
            this.pushQuery({ sql: `alter table ${quotedTableName} add constraint "${constraintName}" ${type.replace(/^text /, '')}`, bindings: [] });
        }
        else {
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} type ${type} using (${colName}::${type})`, bindings: [] });
        }
        that.alterColumnDefault.call(this, col, colName);
        that.alterColumnNullable.call(this, col, colName);
    }
    alterColumnNullable(col, colName) {
        const quotedTableName = this.tableName();
        const nullable = col.modified.nullable;
        if (!nullable) {
            return;
        }
        if (nullable[0] === false) {
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} set not null`, bindings: [] });
        }
        else {
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} drop not null`, bindings: [] });
        }
    }
    alterColumnDefault(col, colName) {
        const quotedTableName = this.tableName();
        const defaultTo = col.modified.defaultTo;
        if (!defaultTo) {
            return;
        }
        if (defaultTo[0] === null) {
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} drop default`, bindings: [] });
        }
        else {
            const modifier = col.defaultTo.apply(col, defaultTo);
            this.pushQuery({ sql: `alter table ${quotedTableName} alter column ${colName} set ${modifier}`, bindings: [] });
        }
    }
}
exports.PostgreSqlConnection = PostgreSqlConnection;
